#include <emma/lib/util.h>
#pragma hdrstop
#include "rexm.h"

static TList<ParamRec>	Params;
static char*	TextStart;
static char*	TextEnd;
static const char*	CodeBase;

static char*	match_node(char*, const char *CodePos);

#define READ(T) ((++(T*)CodePos)[-1])
#define SKIP(n) (((char*)CodePos)+=n)


/*=====================
 Regular expression matcher
=====================*/
boolean PAT_Search(const char*CodePos, char*Text, TList<ParamRec> &Ret)
{
	TextStart=Text;
	TextEnd=Text+strlen(Text);
	CodeBase=CodePos;

	Params=READ(int)+1;	// read the count from the Pattern
	while(Text<TextEnd)
	{
		char* Check;

		if((Check=match_node(Text, CodePos))!=NULL)
		{
			Params[0].L=Text;
			Params[0].R=Check;
			Ret=Params;
			return true;
		}

		Text++;
	}

	return false;
}

boolean PAT_Match(const char*CodePos, char*Text, char*End, TList<ParamRec> &Ret)
{
	TextStart=Text;
	TextEnd=End;
	CodeBase=CodePos;

	Params=READ(int)+1;	// read the count from the Pattern

	char* Check=match_node(Text, CodePos);
	if(!Check)
		return false;

	Params[0].L=Text;
	Params[0].R=Check;
	Ret=Params;
	return true;
}

vstring PAT_Param(ParamRec param)
{
	return vstring(param.L, param.R-param.L);
}

vstring PAT_Describe(const char *Code)
{
	int opcode;
	vstring ret;
	const char *CodePos=Code;

	ret.put_format("SEARCH TEMPLATE ParamCount=%i\n", READ(int));

	while((opcode=READ(char))!=OP_NULL)
	{
		ret.put_format("[%03i] ", CodePos-Code);

		switch(opcode)
		{
		case OP_RET:
			ret.put_format("RET\n");
			break;

		case OP_JUMP:
			ret.put_format("JUMP %i\n", READ(int));
			break;

		case OP_GROUPL:
			ret.put_format("( %i\n", READ(int));
			break;

		case OP_GROUPR:
			ret.put_format(") %i\n", READ(int));
			break;

		case OP_EDGEOFLINE:
			ret.put_format("EDGE OF LINE\n");
			break;

		case OP_NOTEOL:
			ret.put_format("NOT EOL\n");
			break;

		case OP_CHARSET:{
			const char* charset=CodePos;
			ret.put_format("CHARSET [");
			for(int c=0; c<256; c++)
			{
				int t=c >> 3;
				int m=1 << (c & 7);
				if(charset[t] & m)
					ret.put_format("%c", c);
			}
			ret.put_format("]\n");
			SKIP(32);
			break;}

		case OP_PATTERN:{
			int	patlen=READ(int);
			const char* data=CodePos;
			ret.put_format("PATTERN '");
			for(int c=0; c<patlen; c++)
			{
				ret.put_format("%c", data[c]);
			}
			ret.put_format("'\n");
			SKIP(patlen);
			break;}

		case OP_ALT:{
			int t;
			ret.put_format("ALT");
			while((t=READ(int)) != BADI)
			{
				ret.put_format(" CALL(%i)", t);
			}
			ret.put_format("\n");
			break;}

		case OP_RANGE:{
			int	min=READ(int);
			int	max=READ(int);
			boolean	greedy=READ(boolean);
			int nextnode=READ(int);
			ret.put_format("RANGE min=%i max=%i greedy=%i CALL(%i)\n",
				min, max, greedy, nextnode);
			break;}
		}
	}

	ret.put_format("0\n");
	return ret;
}

/*=====================
 Private functions
=====================*/
static char* match_node(char*TextPos, const char *CodePos)
{
	int opcode;

	while((opcode=READ(char)) != OP_RET)
	{
		switch(opcode)
		{
		case OP_JUMP:
			CodePos=CodeBase+READ(int);
			break;

		case OP_GROUPL:
			Params[ READ(int) ].L=TextPos;
			break;

		case OP_GROUPR:
			Params[ READ(int) ].R=TextPos;
			break;

		case OP_EDGEOFLINE:
			if(TextStart<TextPos && TextPos<TextEnd)
			{
				if(TextPos[0]=='\n')
					TextPos++;
				else if(TextPos[0]=='\r' && TextPos[1]=='\n')
					TextPos+=2;
				else return NULL;
			}
			break;

		case OP_NOTEOL:
			if(TextPos>=TextEnd || TextPos[0]=='\n' || TextPos[0]=='\r')
				return NULL;
			TextPos++;
			break;

		case OP_CHARSET:{
			const char* charset=CodePos;
			SKIP(32);

			if(TextPos>=TextEnd)
				return NULL;

			int c=TextPos[0];
			int t=c >> 3;
			int m=1 << (c & 7);
			if((charset[t] & m)==0)
				return NULL;

			TextPos++;
			break;}

		case OP_PATTERN:{
			int	patlen=READ(int);
			const char* Pat=CodePos;
			SKIP(patlen);

			if(TextPos+patlen>TextEnd)
				return NULL;

			for(int i=0; i<patlen; i++)
				if(TextPos[i]!=Pat[i])
					return NULL;
			TextPos+=patlen;
			break;}

		case OP_ALT:{
			int	Disp;
			char*	Check;
			while((Disp=READ(int))!=BADI)
			{
				if((Check=match_node(TextPos, CodeBase+Disp))!=NULL)
					return Check;
			}
			return NULL;}

		case OP_RANGE:{
			int	min=READ(int);
			int	max=READ(int);
			boolean	greedy=READ(boolean);
			const char* nextnode=CodeBase+READ(int);
			const char* fillnode=CodePos;

			int	n;
			char*	Check;

			for(n=0; n < min; n++)
			{
				if((Check=match_node(TextPos, fillnode))==NULL)
					return NULL;

				TextPos=Check;
			}

			char* Found=TextPos;
			for(n=min; n < max; n++)
			{
				if((Check=match_node(TextPos, fillnode))==NULL)
					break;

				TextPos=Check;

				if((Check=match_node(TextPos, nextnode))!=NULL)
				{
					Found=Check;
					if(!greedy)break;
				}
			}
			return Found;}
		}
	}

	return TextPos;
}
